use crate::core::input::InputState;
use crate::core::render::pipelines::triangle_pipeline::shaders::vs;
use crate::core::render::render_context::RenderContext;
use crate::core::scene::Scene;
use crate::core::timer::Timer;
use glam::{Mat3, Mat4, Vec3};
use std::error::Error;
use std::sync::Arc;
use std::time::Instant;
use vulkano::buffer::{Buffer, BufferCreateInfo, BufferUsage, Subbuffer};
use vulkano::command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer};
use vulkano::descriptor_set::{DescriptorSet, WriteDescriptorSet};
use vulkano::memory::allocator::{AllocationCreateInfo, MemoryTypeFilter, StandardMemoryAllocator};
use vulkano::pipeline::{GraphicsPipeline, Pipeline, PipelineBindPoint};
use vulkano_util::renderer::VulkanoWindowRenderer;
use winit::event::ElementState;
use winit::keyboard::{KeyCode, PhysicalKey};

use crate::core::render::pipelines::triangle_pipeline::create_triangle_pipeline;
use crate::core::render::vertex::Vertex2D;

use crate::core::render::vulkan_context::VulkanContext;

const VERTICES: [Vertex2D; 12] = [
    // Triangle en haut à gauche
    Vertex2D {
        position: [-0.5, -0.75],
        color: [1.0, 0.0, 0.0],
    },
    Vertex2D {
        position: [-0.75, -0.25],
        color: [0.0, 1.0, 0.0],
    },
    Vertex2D {
        position: [-0.25, -0.25],
        color: [0.0, 0.0, 1.0],
    },
    // Triangle en bas à gauche
    Vertex2D {
        position: [-0.5, 0.25],
        color: [0.5, 0.5, 0.5],
    },
    Vertex2D {
        position: [-0.75, 0.75],
        color: [0.2, 0.8, 0.2],
    },
    Vertex2D {
        position: [-0.25, 0.75],
        color: [0.8, 0.2, 0.2],
    },
    // Triangle en haut à droite
    Vertex2D {
        position: [0.5, -0.75],
        color: [1.0, 1.0, 0.0],
    },
    Vertex2D {
        position: [0.25, -0.25],
        color: [0.0, 1.0, 1.0],
    },
    Vertex2D {
        position: [0.75, -0.25],
        color: [1.0, 0.0, 1.0],
    },
    // Triangle en bas à droite
    Vertex2D {
        position: [0.5, 0.25],
        color: [0.1, 0.5, 0.8],
    },
    Vertex2D {
        position: [0.25, 0.75],
        color: [0.8, 0.6, 0.1],
    },
    Vertex2D {
        position: [0.75, 0.75],
        color: [0.3, 0.4, 0.6],
    },
];

pub struct MainSceneState {
    pipeline: Arc<GraphicsPipeline>,
    vertex_buffer: Subbuffer<[Vertex2D]>,
    uniform_buffer: Subbuffer<vs::MVPData>,
    rotation: f32,
}

#[derive(Default)]
pub struct MainScene {
    state: Option<MainSceneState>,
}

impl Scene for MainScene {
    fn loaded(&self) -> bool {
        self.state.is_some()
    }

    fn load(&mut self, render_context: &RenderContext) {
        let pipeline =
            create_triangle_pipeline(render_context.device(), render_context.swapchain_format())
                .unwrap();
        let vertex_buffer =
            Vertex2D::create_buffer(Vec::from_iter(VERTICES), render_context.memory_allocator())
                .unwrap();

        let uniform_buffer = MainScene::get_uniform_buffer(
            0.0,
            render_context.memory_allocator(),
            render_context.aspect_ratio(),
        );

        self.state = Some(MainSceneState {
            pipeline,
            vertex_buffer,
            uniform_buffer,
            rotation: 0.0,
        })
    }

    fn update(&mut self, render_context: &RenderContext, input_state: &InputState, timer: &Timer) {
        let state = self.state.as_mut().unwrap();
        let delta_rotation = if input_state.get_key_state(PhysicalKey::Code(KeyCode::KeyA))
            == &ElementState::Pressed
        {
            timer.delta_time() * 5.0
        } else if input_state.get_key_state(PhysicalKey::Code(KeyCode::KeyD))
            == &ElementState::Pressed
        {
            timer.delta_time() * -5.0
        } else {
            timer.delta_time() * 0.0
        };
        state.rotation += delta_rotation;
        state.uniform_buffer = MainScene::get_uniform_buffer(
            state.rotation,
            render_context.memory_allocator(),
            render_context.aspect_ratio(),
        );
    }

    fn render(
        &self,
        render_context: &RenderContext,
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) {
        let vertex_count = self.state.as_ref().unwrap().vertex_buffer.len() as u32;
        let instance_count = vertex_count / 3;

        let layout = &self.state.as_ref().unwrap().pipeline.layout().set_layouts()[0];
        let descriptor_set = DescriptorSet::new(
            render_context.descriptor_set_allocator().clone(),
            layout.clone(),
            [WriteDescriptorSet::buffer(
                0,
                self.state.as_ref().unwrap().uniform_buffer.clone(),
            )],
            [],
        )
        .unwrap();

        unsafe {
            builder
                .bind_pipeline_graphics(self.state.as_ref().unwrap().pipeline.clone())
                .unwrap()
                .bind_descriptor_sets(
                    PipelineBindPoint::Graphics,
                    self.state.as_ref().unwrap().pipeline.layout().clone(),
                    0,
                    descriptor_set,
                )
                .unwrap()
                .bind_vertex_buffers(0, self.state.as_ref().unwrap().vertex_buffer.clone())
                .unwrap()
                .draw(vertex_count, instance_count, 0, 0)
                .unwrap();
        }
    }

    fn unload(&mut self) {}
}

impl MainScene {
    fn get_uniform_buffer(
        rotation: f32,
        memory_allocator: &Arc<StandardMemoryAllocator>,
        aspect_ratio: f32,
    ) -> Subbuffer<vs::MVPData> {
        let rotation = Mat3::from_rotation_y(rotation);

        // NOTE: This teapot was meant for OpenGL where the origin is at the lower left
        // instead the origin is at the upper left in Vulkan, so we reverse the Y axis.
        let proj = Mat4::perspective_rh_gl(std::f32::consts::FRAC_PI_2, aspect_ratio, 0.01, 100.0);
        let view = Mat4::look_at_rh(
            Vec3::new(0.3, 0.3, 1.0),
            Vec3::new(0.0, 0.0, 0.0),
            Vec3::new(0.0, -1.0, 0.0),
        );
        let scale = Mat4::from_scale(Vec3::splat(1.0));

        let uniform_data = vs::MVPData {
            world: Mat4::from_mat3(rotation).to_cols_array_2d(),
            view: (view * scale).to_cols_array_2d(),
            projection: proj.to_cols_array_2d(),
        };

        Buffer::from_data(
            memory_allocator.clone(),
            BufferCreateInfo {
                usage: BufferUsage::UNIFORM_BUFFER,
                ..Default::default()
            },
            AllocationCreateInfo {
                memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                    | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
                ..Default::default()
            },
            uniform_data,
        )
        .unwrap()
    }
}
